// BenLincoln.TheLostWorlds.CDDigitalAudio
// Copyright 2006-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.TheLostWorlds.CDDigitalAudio.

// BenLincoln.TheLostWorlds.CDDigitalAudio is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.TheLostWorlds.CDDigitalAudio is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.TheLostWorlds.CDDigitalAudio (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Text;
using BLCDDA = BenLincoln.TheLostWorlds.CDDigitalAudio;
using BLDA = BenLincoln.DigitalAudio;
using BLS = BenLincoln.Shared;

namespace BenLincoln.TheLostWorlds.CDDigitalAudio
{
    public class SR1PCMonolithicAudioFile : BLCDDA.SoundClipContainerFile
    {
        protected enum ScanState
        {
            LookingForFirstClip,
            LookingForBeginningOfZeroes,
            LookingForEndOfZeroes,
            FoundZeroes
        }

        // how many 0x00 bytes in a row indicates a break between clips?
        protected int _ZeroByteThreshold = 128;

        // how many bytes long must a clip be to be put in the list of clips for the file?
        protected int _ClipLengthThreshold = 1;

        // how many channels do the audio files inside have?
        protected ushort _NumChannels;

        public SR1PCMonolithicAudioFile(string path, ushort numChannels)
            : this(path, numChannels, 128, 1)
        {
        }

        public SR1PCMonolithicAudioFile(string path, ushort numChannels, int zeroByteThreshold, int clipLengthThreshold)
            : base(path)
        {
            _NumChannels = numChannels;
            _ZeroByteThreshold = zeroByteThreshold;
            _ClipLengthThreshold = clipLengthThreshold;
        }

        public override void LoadFile()
        {
            // scan through the file and figure out where the clips are
            FileStream scanStream = new FileStream(_Path, FileMode.Open, FileAccess.Read);
            int zeroCount = 0;
            long zeroStart = 0;
            long currentFileStart = 0;
            long currentPosition = 0;
            long streamLength = scanStream.Length;
            _IOPercent = 0;
            ArrayList clipList = new ArrayList();

            bool atEndOfStream = false;

            ScanState scanState = ScanState.LookingForFirstClip;

            do
            {
                int currentByte = scanStream.ReadByte();
                if (currentByte == -1)
                {
                    atEndOfStream = true;
                }
                else
                {
                    switch (scanState)
                    {
                        case ScanState.LookingForFirstClip:
                            if (currentByte != 0)
                            {
                                currentFileStart = currentPosition;
                                scanState = ScanState.LookingForBeginningOfZeroes;
                            }
                            break;
                        case ScanState.LookingForBeginningOfZeroes:
                            if (currentByte == 0)
                            {
                                zeroCount = 0;
                                zeroStart = currentPosition;
                                // zero position must start on an odd byte number
                                // so that the clip is an even number of bytes
                                //if ((zeroStart % 2) == 0)
                                //{
                                //    zeroStart++;
                                //}
                                scanState = ScanState.LookingForEndOfZeroes;
                            }
                            break;
                        case ScanState.LookingForEndOfZeroes:
                            if (currentByte == 0)
                            {
                                zeroCount++;
                            }
                            else
                            {
                                scanState = ScanState.FoundZeroes;
                            }
                            break;
                        case ScanState.FoundZeroes:
                            if (zeroCount < _ZeroByteThreshold)
                            {
                                // not enough zeroes to do anything
                                // reset counter and resume scanning
                                zeroCount = 0;
                                scanState = ScanState.LookingForBeginningOfZeroes;
                            }
                            else
                            {
                                // if the length of the found clip is greater than the threshold,
                                // add it to the list
                                long currentLength = (zeroStart + zeroCount) - currentFileStart + 1;
                                // clip length must be even
                                if ((currentLength % 2) == 1)
                                {
                                    currentLength++;
                                } 
                                if (currentLength > _ClipLengthThreshold)
                                {
                                    BLS.OffsetAndLength currentFile = new BLS.OffsetAndLength(
                                        currentFileStart, currentLength);
                                    clipList.Add(currentFile);
                                    // set the beginning of the next file to the next byte
                                    currentFileStart = currentPosition;
                                    // file must start on an even byte number
                                    if ((currentFileStart % 2) == 1)
                                    {
                                        currentFileStart--;
                                    }
                                }
                                // reset counter and resume scanning
                                zeroCount = 0;
                                scanState = ScanState.LookingForBeginningOfZeroes;
                            }
                            break;
                    }
                }
                _IOPercent = (double)currentPosition / (double)streamLength;
                currentPosition++;
            } while (!atEndOfStream);
            scanStream.Close();

            // get the last clip, if there was one
            if (scanState == ScanState.LookingForBeginningOfZeroes)
            {
                zeroCount = 0;
                zeroStart = currentPosition;
                // must start on an even byte number
                if ((zeroStart % 2) == 1)
                {
                    zeroStart++;
                }
                scanState = ScanState.LookingForEndOfZeroes;
            }
            if (scanState == ScanState.LookingForEndOfZeroes)
            {
                // if the length of the found clip is greater than the threshold,
                // add it to the list
                long currentLength = zeroStart - currentFileStart;
                if (currentLength > _ClipLengthThreshold)
                {
                    BLS.OffsetAndLength currentFile = new BLS.OffsetAndLength(
                        currentFileStart, currentLength);
                    clipList.Add(currentFile);
                }
            }

            string[] pathSplit = _Path.Split('\\');
            string fileName = pathSplit[pathSplit.GetUpperBound(0)];
            fileName = fileName.Substring(0, fileName.LastIndexOf('.'));

            // read the arraylist into the array
            _SoundClips = new BLCDDA.SoundClip[clipList.Count];
            int clipNum = 0;
            foreach (BLS.OffsetAndLength oal in clipList)
            {
                _SoundClips[clipNum] = new BLCDDA.SR1PCMonolithicAudioClip(fileName + string.Format("-{0:0000}", clipNum), this, oal.Offset, oal.Length, _NumChannels);
                clipNum++;
            }

            _IOPercent = 1;
        }

        protected override Hashtable GetSubclassInfo()
        {
            Hashtable info = new Hashtable();
            info.Add("Type", "Soul Reaver (PC) Monolithic Sound Clip Collection");
            return info;
        }
    }
}
