// BenLincoln.TheLostWorlds.CDDigitalAudio
// Copyright 2006-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.TheLostWorlds.CDDigitalAudio.

// BenLincoln.TheLostWorlds.CDDigitalAudio is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.TheLostWorlds.CDDigitalAudio is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.TheLostWorlds.CDDigitalAudio (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Text;
using BLCDDA = BenLincoln.TheLostWorlds.CDDigitalAudio;
using BLDA = BenLincoln.DigitalAudio;
using BLS = BenLincoln.Shared;

namespace BenLincoln.TheLostWorlds.CDDigitalAudio
{
    public class SoundClip : IDisposable, BLS.IProgress
    {
        // defined in subclasses
        protected string _TypeName;

        // name of the file
        protected string _Name;

        // parent file which contains this clip
        protected BLCDDA.SoundClipContainerFile _ParentSoundClipContainerFile;

        // offset in the parent file
        protected long _Offset;

        // length in bytes in the parent file
        protected long _Length;

        // 22050, 44100, etc
        protected uint _SampleRate;

        // 8, 16, etc.
        protected ushort _BitDepth;

        // a positive integer
        protected ushort _NumChannels;

        // loop points
        protected BLDA.LoopPoint[] _LoopPoints;

        // a temp wav file to store the uncompressed version in
        protected BLDA.WAVFile _WAVFile;

        // read/write progress
        protected double _IOPercent = 0;

        // error during operation threaded operation?
        protected bool _ErrorOccurred = false;

        // what was the error message?
        protected string _LastErrorMessage = "";

        #region Properties

        public string TypeName
        {
            get
            {
                return _TypeName;
            }
        }

        public string Name
        {
            get
            {
                return _Name;
            }
        }

        public BLCDDA.SoundClipContainerFile ParentSoundClipContainerFile
        {
            get
            {
                return _ParentSoundClipContainerFile;
            }
        }

        public long Offset
        {
            get
            {
                return _Offset;
            }
        }

        public long Length
        {
            get
            {
                return _Length;
            }
        }

        public uint SampleRate
        {
            get
            {
                return _SampleRate;
            }
        }

        public ushort BitDepth
        {
            get
            {
                return _BitDepth;
            }
        }

        public ushort NumChannels
        {
            get
            {
                return _NumChannels;
            }
        }

        public BLDA.LoopPoint[] LoopPoints
        {
            get
            {
                return _LoopPoints;
            }
        }

        public BLDA.WAVFile WAVFile
        {
            get
            {
                return _WAVFile;
            }
        }

        public bool ErrorOccurred
        {
            get
            {
                return _ErrorOccurred;
            }
        }

        public string LastErrorMessage
        {
            get
            {
                return _LastErrorMessage;
            }
        }

        #endregion

        // for types that have the length in the header at the passed offset
        public SoundClip(string name, BLCDDA.SoundClipContainerFile parentFile, long offset)
            : this(name, parentFile, offset, 0)
        {
        }

        // for types that have the length in the global file header
        public SoundClip(string name, BLCDDA.SoundClipContainerFile parentFile, long offset, long length)
        {
            _Name = name;
            _ParentSoundClipContainerFile = parentFile;
            _Offset = offset;
            _Length = length;
            _IOPercent = 0;
            _IOPercent = 1;
        }

        public virtual void LoadFileData()
        {
        }

        // get a hashtable of all the global properties of the file
        public Hashtable GetInfo()
        {
            Hashtable info = GetSubclassInfo();
            //info.Add("Name", _Name);
            info.Add("Offset in file", _Offset.ToString());
            info.Add("Length", _Length.ToString() + " bytes");
            info.Add("Sample Rate", _SampleRate.ToString() + " Hz");
            info.Add("Bit Depth", _BitDepth.ToString() + "-bit");
            info.Add("Channels", _NumChannels.ToString());
            if (_LoopPoints != null)
            {
                int numLoops = _LoopPoints.GetLength(0);
                for (int i = 0; i < numLoops; i++)
                {
                    info.Add("Loop " + string.Format("{0:00}", i + 1),
                        "Start: " + _LoopPoints[i].Start.ToString() +
                        " End: " + _LoopPoints[i].End.ToString());
                }
            }
            return info;
        }

        // subclasses only have to define things that are unique to them
        protected virtual Hashtable GetSubclassInfo()
        {
            Hashtable info = new Hashtable();
            info.Add("Type", "Sound Clip Superclass");
            return info;
        }

        // shouldn't ever need to override this, but who knows?
        public virtual void ExportFile(string outPath)
        {
            _ErrorOccurred = false;
            _LastErrorMessage = "";
            try
            {
                FileStream outStream = new FileStream(outPath, FileMode.Create, FileAccess.Write);
                _WAVFile.WriteData(outStream);
                outStream.Close();
            }
            catch (Exception ex)
            {
                _ErrorOccurred = true;
                _LastErrorMessage = ex.Message;
            }
        }
        // for threaded exports
        public void ExportFileThreaded(object outPath)
        {
            try
            {
                ExportFile((string)outPath);
            }
            catch (InvalidCastException ex)
            {
                throw new InvalidCastException("Passed object must be a string");
            }
        }

        public double Progress()
        {
            return _IOPercent;
        }

        public void Dispose()
        {
            Dispose(true);
        }

        protected void Dispose(bool disposing)
        {
            if (disposing && (_WAVFile != null))
            {
                _WAVFile.Dispose();
            }
        }
    }
}
