// The Eye of RAW
// Copyright 2006-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of The Eye of RAW.

// The Eye of RAW is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// The Eye of RAW is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with The Eye of RAW (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using CDI = BenLincoln.TheLostWorlds.CDImages;

namespace EyeOfRAW
{
    public partial class frmMain : Form
    {
        protected CDI.RAWImage currentFile;

        #region Constants

        protected const int _MinimumWindowWidth = 300;
        protected const int _MinimumWindowHeight = 200;

        #endregion

        public frmMain(string[] args)
        {
            InitializeComponent();

            string[] commandLineParms = args;

            if (commandLineParms.GetUpperBound(0) > -1)
            {
                string imagePath = args[0];
                if (File.Exists(imagePath))
                {
                    OpenFile(imagePath);
                }
                else
                {
                    MessageBox.Show("Unable to find the file " + imagePath, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
        }

        private void mnuOpen_Click(object sender, EventArgs e)
        {
            OpenFileDialog oDialogue;
            DialogResult result;
            oDialogue = new OpenFileDialog();
            oDialogue.CheckPathExists = true;
            oDialogue.CheckFileExists = true;
            oDialogue.Multiselect = false;
            oDialogue.Title = "Select a RAW or SRTex file to open...";
            oDialogue.Filter = "RAW and SRTex Images (*.RAW; *.SRTex)|*.RAW; *.SRTex";
            result = oDialogue.ShowDialog();
            if (result == DialogResult.OK)
            {
                OpenFile(oDialogue.FileName);
            }

        }

        protected void OpenFile(string fileName)
        {
            try
            {
                CDI.RAWImage image = CDI.RAWImage.FromFile(fileName);
                imgDisplay.Image = image.GetBitmap();
                reconfigureWindow(image.Width, image.Height, Path.GetFileName(image.FilePath));
                currentFile = image;
                //currentFile = new ImageFile(fileName);
                //mnuProperties.Enabled = true;
                //currentFile.ReadImageFromFile();
                //imgDisplay.Image = currentFile.Image;
                //reconfigureWindow(currentFile.ImageWidth, currentFile.ImageHeight, currentFile.FileName);

                mnuSaveAsPNG.Enabled = true;
            }
            catch (UnrecognizedFileTypeException uEx)
            {
                MessageBox.Show("Eye of RAW does not currently support this file format.", "Unsupported File Type",
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            catch (Exception ex)
            {
                MessageBox.Show("Eye of RAW was unable to read this file. The following exception was thrown: " + ex.Message, 
                    "File I/O Error",
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void mnuSaveAsPNG_Click(object sender, EventArgs e)
        {
            SaveFileDialog sDialogue;
            DialogResult result;
            sDialogue = new SaveFileDialog();
            sDialogue.Filter = "Portable Network Graphic Images (*.PNG)|*.PNG";
            //remove extension
            string fnt = Path.GetFileName(currentFile.FilePath);
            string fName = fnt.Substring(0, fnt.LastIndexOf('.'));
            sDialogue.FileName = fName + ".PNG";
            sDialogue.DefaultExt = "PNG";
            sDialogue.AddExtension = true;
            result = sDialogue.ShowDialog();
            if (result == DialogResult.OK)
            {
                imgDisplay.Image.Save(sDialogue.FileName, System.Drawing.Imaging.ImageFormat.Png);
            }           
        }

        private void mnuProperties_Click(object sender, EventArgs e)
        {
            frmProperties PropertiesDialogue = new frmProperties();
            if (currentFile != null)
            {
                PropertiesDialogue.UpdateInformation(currentFile);
            }
            PropertiesDialogue.ShowDialog();
            PropertiesDialogue.Dispose();
        }

        private void mnuExit_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        private void mnuOptions_Click(object sender, EventArgs e)
        {
            ProgramSettings currentSettings = new ProgramSettings();
            frmOptions OptionsDialogue = new frmOptions(currentSettings);
            OptionsDialogue.ShowDialog();
            OptionsDialogue.Dispose();
        }

        private void mnuAbout_Click(object sender, EventArgs e)
        {
            frmAbout AboutDialogue = new frmAbout();
            AboutDialogue.ShowDialog();
            AboutDialogue.Dispose();
        }

        #region File I/O

        #endregion

        #region Options/Registry

        //protected void UpdateOptionsFromWindow(frmOptions optionsDialogue)
        //{
        //    ProgramSettings newSettings = new ProgramSettings();
        //    newSettings.IgnoreAlpha = optionsDialogue.IgnoreAlpha;
        //    newSettings.WriteOptionsToRegistry();
        //}

        #endregion
        #region Utility Functions

        protected void reconfigureWindow(int imageWidth, int imageHeight, string newTitle)
        {
            this.Width = Math.Max(_MinimumWindowWidth,
                Math.Min((32 + imageWidth), System.Windows.Forms.Screen.GetBounds(new Point(0, 0)).Width));
            this.Height = Math.Max(_MinimumWindowHeight,
                Math.Min((72 + imageHeight), System.Windows.Forms.Screen.GetBounds(new Point(0, 0)).Height));
            this.Text = "The Eye of RAW - " + newTitle;
        }

        protected uint ReadMask(BinaryReader reader)
        {
            uint byte1 = reader.ReadByte();
            uint byte2 = reader.ReadByte();
            uint byte3 = reader.ReadByte();
            uint byte4 = reader.ReadByte();

            byte1 <<= 24;
            byte2 <<= 16;
            byte3 <<= 8;

            return byte1 | byte2 | byte3 | byte4;
        }

        protected uint TransformMask(uint inMask)
        {
            uint byte1 = inMask & 0xFF;
            uint byte2 = inMask & 0xFF00;
            uint byte3 = inMask & 0xFF0000;
            uint byte4 = inMask & 0xFF000000;

            byte1 <<= 24;
            byte2 <<= 8;
            byte3 >>= 8;
            byte4 >>= 24;

            return byte1 | byte2 | byte3 | byte4;
        }

        protected int getBitPositionInDWord(uint inDWord)
        {
            uint bitMask, masked;

            for (int i = 0; i < 32; i++)
            {
                bitMask = (uint)(1 << (31 - i));
                masked = (inDWord & bitMask) >> (31 - i);
                if (masked == 1)
                {
                    return i;
                }
            }
            return 0;
        }

        protected int getBitPositionInWord(uint inWord)
        {
            uint bitMask, masked;

            for (int i = 0; i < 16; i++)
            {
                bitMask = (ushort)(1 << (15 - i));
                masked = (inWord & bitMask) >> (15 - i);
                if (masked == 1)
                {
                    return i;
                }
            }
            return 0;
        }

    #endregion
    }
}