// The Sound of Her Wings
// Copyright 2006-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of The Sound of Her Wings.

// The Sound of Her Wings is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// The Sound of Her Wings is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with The Sound of Her Wings (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using Microsoft.Win32;
using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.IO;
using System.Media;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using DX = Microsoft.DirectX;
using DXAVP = Microsoft.DirectX.AudioVideoPlayback;
using BLCDDA = BenLincoln.TheLostWorlds.CDDigitalAudio;
using BLDA = BenLincoln.DigitalAudio;
using BLS = BenLincoln.Shared;
using BLUI = BenLincoln.UI;

namespace TheSoundOfHerWings
{
    public partial class frmMain : Form, BLS.IProgress
    {
        // enums
        public enum FileType
        {
            WAV,
            SR1MusicDotDat,
            SR1VoicesDotDat,
            SR1PCPMF
        }

        #region class variables

        // temp files for wav channels
        //protected BLS.TempFile _WaveTempFile;
        //protected BLS.TempFile[] _ChannelTempFiles;
        protected string[] _ChannelTempFilePaths;

        // the file that is currently open, and the clip that's open from within that file
        protected BLCDDA.SoundClipContainerFile _CurrentFile;
        protected BLCDDA.SoundClip _CurrentClip;

        // waveform displays
        protected BLDA.WaveformDisplay[] _WaveformDisplays;

        // display parameters
        Color _WFBackgroundColour = Color.FromArgb(255, 128, 128, 128);
        Color _WFGridColour = Color.FromArgb(128, 64, 64, 64);
        //Color _WFWaveformColour = Color.FromArgb(220, 0, 255, 0);
        //Color _WFWaveformColour = Color.FromArgb(190, 196, 0, 255);
        //Color _WFWaveformColour = Color.FromArgb(190, 140, 0, 255);   // initial wave colour

        //// purple (default - original)
        //Color _WFWaveformColour = Color.FromArgb(160, 100, 0, 255);
        //Color _WFWaveformMedianColour = Color.FromArgb(190, 255, 0, 255);

        // purple (default)
        Color _WFWaveformColour = Color.FromArgb(160, 100, 0, 255);
        Color _WFWaveformMedianColour = Color.FromArgb(190, 255, 24, 255);

        Color _WFXAxisColour = Color.FromArgb(64, 220, 220, 220);
        //Color _WFXAxisColour = Color.FromArgb(64, 64, 64, 220);
        Color _WFCursorColour = Color.FromArgb(128, 24, 24, 255);

        // waveform drawing options
        protected bool _UseGradientDrawMode = true;
        protected bool _DrawWaveform = true;

        // playback options
        protected int _LoopIterations = 1;

        // export options
        protected bool _ExportLoopDataAsText = false;

        // height of the time display
        protected const int _TimeDisplayHeight = 30;

        // DX Audio playback object
        protected DXAVP.Audio _Audio;

        // timer object for keeping the position cursor(s) updated
        protected System.Windows.Forms.Timer _Timer;

        // progress window
        protected BLUI.ProgressWindow progressWindow;

        // progress value for this window
        protected double _Progress;

        // registry preferences
        protected BLS.RegistryPreferenceManager _PrefManager;

        // preferences window
        protected frmPreferences prefsWindow;

        // about window
        protected frmAbout aboutWindow;

        // commandline arguments
        protected string[] _Args;

        // last folders used
        protected string _LastOpenFolder = "";
        protected string _LastExportFolder = "";
        protected string _LastExportAllFolder = "";

        // start time for wav (for clicking on the waveform display while stopped)
        protected double _WaveformStartPercent = 0;

        // current loop number within the clip
        protected int _CurrentLoop = 0;

        // number of times the current loop has been played
        protected int _LoopCount = 0;

        // loop start and end time as percentages
        protected double _CurrentLoopStart = 0;
        protected double _CurrentLoopEnd = 0;

        // last index in the clip list, to prevent reload of waves when one really hasn't happened
        protected int _LastClipListIndex = -1;

        #endregion

        public frmMain(string[] args)
        {
            InitializeComponent();

            _PrefManager = new BLS.RegistryPreferenceManager("Software\\The Lost Worlds\\The Sound of Her Wings");
            ReadOptionsFromRegistry();

            CreateInitialWaveformDisplay();

            // set up the tooltips
            CreateToolTip(this.btnOpen, "Open a sound file");
            CreateToolTip(this.btnExport, "Export the currently-selected sound clip");
            CreateToolTip(this.btnExportAll, "Export all sound clips in the source file");
            CreateToolTip(this.btnPlay, "Starts or resumes playback");
            CreateToolTip(this.btnPause, "Pauses playback");
            CreateToolTip(this.btnStop, "Stops playback");

            // disable controls that aren't applicable until a file is opened
            EnableExportControls(false);
            EnableExportAllControls(false);
            EnablePlaybackControls(false);

            _Args = args;
        }

        private void frmMain_Load(object sender, EventArgs e)
        {
            // if an argument was passed, try to open the file
            if (_Args.GetUpperBound(0) > -1)
            {
                string filePath = _Args[0];
                if (File.Exists(filePath))
                {
                    LoadFile(filePath);
                }
                else
                {
                    MessageBox.Show("Unable to find the file " + filePath, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
        }

        #region File import

        protected void OpenFile()
        {
            OpenFileDialog openFile;
            DialogResult result;
            openFile = new OpenFileDialog();
            openFile.CheckFileExists = true;
            openFile.CheckPathExists = true;
            openFile.Title = "Select a file to open...";
            //openFile.FileName = "bigfile.dat";
            openFile.DefaultExt = "*.*";
            //openFile.Filter = "WAV Files (*.WAV)|*.WAV|" +
            //    "Soul Reaver PC Music (music.dat)|music.dat|" +
            //    "Soul Reaver PC Voices (voices.dat)|voices.dat|" +
            //    "All Files (*.*)|*.*";
            openFile.Filter = BuildFileFilter();

            // use last used folder if it exists
            if ((_LastOpenFolder != "") && (Directory.Exists(_LastOpenFolder)))
            {
                openFile.InitialDirectory = _LastOpenFolder;
            }

            result = openFile.ShowDialog();
            if (result == DialogResult.OK)
            {
                // update the last used folder
                _LastOpenFolder = GetDirectoryNameFromPath(openFile.FileName);
                WriteStringPrefsToRegistry(GetCurrentStringPrefs());
                LoadFile(openFile.FileName);
            }
        }

        protected string BuildFileFilter()
        {
            ArrayList fileTypes = GetSupportedFileTypes();
            string filter = "All Supported File Types|";
            foreach (BLS.FileTypeAndExtension type in fileTypes)
            {
                filter += type.FileFilter + ";";
            }
            // remove the last comma
            filter = filter.Substring(0, filter.Length - 1) + "|";

            // add each of the types individually
            foreach (BLS.FileTypeAndExtension type in fileTypes)
            {
                filter += type.FileType + " (" + type.FileFilter + ")|" + type.FileFilter + "|";
            }

            // add the "all files" option
            filter += "All Files (*.*)|*.*";

            return filter;
        }

        protected ArrayList GetSupportedFileTypes()
        {
            ArrayList types = new ArrayList();

            BLS.FileTypeAndExtension wav = new BLS.FileTypeAndExtension
                (
                "RIFF WAVE", "wav", "*.WAV"
                );
            types.Add(wav);
            BLS.FileTypeAndExtension sr1music = new BLS.FileTypeAndExtension
                (
                "Soul Reaver (PC) Music", "dat", "music.dat"
                );
            types.Add(sr1music);
            BLS.FileTypeAndExtension sr1voices = new BLS.FileTypeAndExtension
                (
                "Soul Reaver (PC) Voices", "dat", "voices.dat"
                );
            types.Add(sr1voices);
            BLS.FileTypeAndExtension sr1pmf = new BLS.FileTypeAndExtension
                (
                "Soul Reaver (PC) Object Sound Collection", "pmf", "*.pmf"
                );
            types.Add(sr1pmf);

            return types;
        }

        protected void LoadFile(string filePath)
        {
            // if there are temp files left over from a previous file, delete them
            DeleteTempFiles();
            // clean up any previous file
            if (_CurrentFile != null)
            {
                _CurrentFile.Dispose();
            }

            // determine what load method to use
            FileType type = FileType.WAV;

            string fileName = GetFileNameFromPath(filePath);
            string fileExt = fileName.Substring(fileName.LastIndexOf('.'));
            if (fileExt.ToLower() == ".wav")
            {
                type = FileType.WAV;
            }
            else if (fileExt.ToLower() == ".pmf")
            {
                type = FileType.SR1PCPMF;
            }
            else if (fileName.ToLower() == "music.dat")
            {
                type = FileType.SR1MusicDotDat;
            }
            else if (fileName.ToLower() == "voices.dat")
            {
                type = FileType.SR1VoicesDotDat;
            }
            else
            {
                MessageBox.Show("Unable to determine file type.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            //ReadWAVFile(openFile.FileName);
            Thread loadThread = new Thread(ReadWAVFileThreaded);

            switch (type)
            {
                case FileType.WAV:
                    loadThread = new Thread(ReadWAVFileThreaded);
                    break;
                case FileType.SR1MusicDotDat:
                    loadThread = new Thread(ReadSR1MusicDotDatThreaded);
                    break;
                case FileType.SR1VoicesDotDat:
                    loadThread = new Thread(ReadSR1VoicesDotDatThreaded);
                    break;
                case FileType.SR1PCPMF:
                    loadThread = new Thread(ReadSR1PCPMFThreaded);
                    break;
            }

            loadThread.Start(filePath);
        }

        protected void ReadWAVFileThreaded(object filePath)
        {
            try
            {
                ReadWAVFile((string)filePath);
            }
            catch (InvalidCastException ex)
            {
                throw new InvalidCastException("Passed object must be a string");
            }
        }

        protected void ReadWAVFile(string filePath)
        {
            _CurrentFile = new BLCDDA.WAVContainer(filePath);
            _CurrentFile.LoadFile();    // don't need to multithread this for WAV files

            CallThreadedClipLoad();

            Invoke(new MethodInvoker(DoPostFileLoadUISetup));
        }

        protected void ReadSR1MusicDotDatThreaded(object filePath)
        {
            try
            {
                ReadSR1MusicDotDat((string)filePath);
            }
            catch (InvalidCastException ex)
            {
                throw new InvalidCastException("Passed object must be a string");
            }
        }

        protected void ReadSR1MusicDotDat(string filePath)
        {
            _CurrentFile = new BLCDDA.SR1PCMonolithicAudioFile(filePath, 2, 128, 44100);
            CallThreadedFileLoad();
            CallThreadedClipLoad();

            Invoke(new MethodInvoker(DoPostFileLoadUISetup));
        }

        protected void ReadSR1VoicesDotDatThreaded(object filePath)
        {
            try
            {
                ReadSR1VoicesDotDat((string)filePath);
            }
            catch (InvalidCastException ex)
            {
                throw new InvalidCastException("Passed object must be a string");
            }
        }

        protected void ReadSR1VoicesDotDat(string filePath)
        {
            //_CurrentFile = new BLCDDA.SR1PCMonolithicAudioFile(filePath, 1, 256, 1);
            _CurrentFile = new BLCDDA.SR1PCMonolithicAudioFile(filePath, 1, 168, 16384);
            CallThreadedFileLoad();
            CallThreadedClipLoad();

            Invoke(new MethodInvoker(DoPostFileLoadUISetup));
        }

        protected void ReadSR1PCPMFThreaded(object filePath)
        {
            try
            {
                ReadSR1PCPMF((string)filePath);
            }
            catch (InvalidCastException ex)
            {
                throw new InvalidCastException("Passed object must be a string");
            }
        }

        protected void ReadSR1PCPMF(string filePath)
        {
            _CurrentFile = new BLCDDA.SR1PCPMFFile(filePath);
            CallThreadedFileLoad();
            CallThreadedClipLoad();

            Invoke(new MethodInvoker(DoPostFileLoadUISetup));
        }

        protected void CallThreadedFileLoad()
        {
            Invoke(new MethodInvoker(CreateProgressWindow));
            progressWindow.SetTitle("Loading");
            progressWindow.SetMessage("Scanning file");
            Thread scanThread = new Thread(_CurrentFile.LoadFile);
            scanThread.Start();
            // wait for the loading to finish, and keep the load value in this class updated
            do
            {
                progressWindow.SetProgress((int)(_CurrentFile.Progress() * (double)100));
                Thread.Sleep(5);
            }
            while (scanThread.IsAlive);
            Invoke(new MethodInvoker(DestroyProgressWindow));
        }

        protected void CallThreadedClipLoad()
        {
            Invoke(new MethodInvoker(CreateProgressWindow));
            progressWindow.SetTitle("Loading");
            progressWindow.SetMessage("Loading audio data");
            Thread loadThread = new Thread(_CurrentFile.LoadClipData);
            loadThread.Start();
            // wait for the loading to finish, and keep the load value in this class updated
            do
            {
                progressWindow.SetProgress((int)(_CurrentFile.Progress() * (double)100));
                Thread.Sleep(5);
            }
            while (loadThread.IsAlive);
            Invoke(new MethodInvoker(DestroyProgressWindow));
        }

        #endregion

        #region File export

        protected void ExportCurrent()
        {
            SaveFileDialog sDialogue;
            DialogResult result;

            if (_CurrentFile == null)
            {
                MessageBox.Show("No file is currently open.",
                    "Export Error",
                    MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
                return;
            }

            if (_CurrentClip == null)
            {
                MessageBox.Show("No clip is currently open.",
                    "Export Error",
                    MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
                return;
            }

            if (lbClipList.SelectedIndex == -1)
            {
                MessageBox.Show("No clip is currently selected.",
                    "Export Error",
                    MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
                return;
            }

            sDialogue = new SaveFileDialog();
            sDialogue.AddExtension = true;
            sDialogue.DefaultExt = "*.wav";
            sDialogue.Filter = "RIFF WAV Files (*.wav)|*.wav";
            sDialogue.FileName = _CurrentClip.Name;
            sDialogue.OverwritePrompt = true;
            sDialogue.Title = "Export Sound Clip as WAV...";
            // use last used folder if it exists
            if ((_LastExportFolder != "") && (Directory.Exists(_LastExportFolder)))
            {
                sDialogue.InitialDirectory = _LastExportFolder;
            }

            result = sDialogue.ShowDialog();

            if (result == DialogResult.OK)
            {
                // update the last used folder
                _LastExportFolder = GetDirectoryNameFromPath(sDialogue.FileName);
                WriteStringPrefsToRegistry(GetCurrentStringPrefs());
                Thread exportThread = new Thread(ExportCurrentFileThreaded);
                exportThread.Start(sDialogue.FileName);
            }
        }

        protected void ExportCurrentFileThreaded(object filePath)
        {
            try
            {
                ExportCurrentFile((string)filePath);
            }
            catch (InvalidCastException ex)
            {
                throw new InvalidCastException("Passed object must be a string");
            }
        }

        protected void ExportCurrentFile(string filePath)
        {
            Invoke(new MethodInvoker(CreateProgressWindow));
            progressWindow.SetTitle("Exporting");
            progressWindow.SetMessage("Exporting WAV");
            Thread exportThread = new Thread(_CurrentClip.ExportFileThreaded);
            exportThread.Start(filePath);
            // wait for the loading to finish, and keep the load value in this class updated
            do
            {
                progressWindow.SetProgress((int)(_CurrentClip.Progress() * (double)100));
                Thread.Sleep(5);
            }
            while (exportThread.IsAlive);
            string filePathDirectory = GetDirectoryNameFromPath(filePath);
            if (_CurrentClip.ErrorOccurred)
            {
                DialogResult result;
                result = MessageBox.Show("An error occurred during the export of file " +
                    filePath + "." + Environment.NewLine +
                    "The error message was: " + _CurrentClip.LastErrorMessage, "Export Error",
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            else
            {
                if (_ExportLoopDataAsText)
                {
                    WriteLoopDataToFile(_CurrentClip, filePathDirectory + _CurrentClip.Name + "-LoopPoints.txt");
                }
            }
            Invoke(new MethodInvoker(DestroyProgressWindow));
        }



        protected void ExportAll()
        {
            FolderBrowserDialog fDialogue;
            DialogResult result1, result2;

            if (_CurrentFile == null)
            {
                MessageBox.Show("No file is currently open.",
                    "Export Error",
                    MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
                return;
            }

            if (_CurrentFile.SoundClips.GetLength(0) == 0)
            {
                MessageBox.Show("The file that is open does not contain any sound clips.",
                    "Export Error",
                    MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
                return;
            }

            fDialogue = new FolderBrowserDialog();
            fDialogue.ShowNewFolderButton = true;
            // use last used folder if it exists
            if ((_LastExportAllFolder != "") && (Directory.Exists(_LastExportAllFolder)))
            {
                fDialogue.SelectedPath = _LastExportAllFolder;
            }

            result1 = fDialogue.ShowDialog();

            if (result1 == DialogResult.OK)
            {
                result2 = MessageBox.Show(
                    "Any files in the selected folder with names identical to those being exported will be overwritten." +
                    Environment.NewLine + 
                    "Are you sure you want to proceed?", "Export", MessageBoxButtons.YesNo, MessageBoxIcon.Question);
                if (result2 == DialogResult.Yes)
                {
                    // update the last used folder
                    _LastExportAllFolder = fDialogue.SelectedPath + "\\";
                    WriteStringPrefsToRegistry(GetCurrentStringPrefs());
                    Thread exportThread = new Thread(ExportAllFilesThreaded);
                    exportThread.Start(fDialogue.SelectedPath + "\\");
                }
            }
        }

        protected void ExportAllFilesThreaded(object basePath)
        {
            try
            {
                ExportAllFiles((string)basePath);
            }
            catch (InvalidCastException ex)
            {
                throw new InvalidCastException("Passed object must be a string");
            }
        }

        protected void ExportAllFiles(string basePath)
        {
            Invoke(new MethodInvoker(CreateProgressWindow));
            progressWindow.SetTitle("Exporting All Sound Clips");
            int clipNum = 1;
            foreach (BLCDDA.SoundClip clip in _CurrentFile.SoundClips)
            {
                progressWindow.SetMessage("Exporting WAV " + clipNum.ToString() + " of " + _CurrentFile.SoundClips.GetLength(0).ToString());
                Thread exportThread = new Thread(clip.ExportFileThreaded);
                exportThread.Start(basePath + clip.Name + ".wav");
                // wait for the loading to finish, and keep the load value in this class updated
                do
                {
                    //progressWindow.SetProgress((int)(clip.Progress() * (double)100));
                    progressWindow.SetProgress((int)((double)(clipNum - 1) / (double)_CurrentFile.SoundClips.GetLength(0) * (double)100));
                    Thread.Sleep(5);
                }
                while (exportThread.IsAlive);
                if (clip.ErrorOccurred)
                {
                    DialogResult result;
                    if (clipNum != _CurrentFile.SoundClips.GetLength(0))
                    {
                        result = MessageBox.Show("An error occurred during the export of file " +
                            basePath + clip.Name + ".wav" + "." + Environment.NewLine +
                            "The error message was: " + clip.LastErrorMessage + Environment.NewLine +
                            "Do you want to continue exporting the other files?", "Export Error",
                            MessageBoxButtons.YesNo, MessageBoxIcon.Error);
                    }
                    else
                    {
                        result = MessageBox.Show("An error occurred during the export of file " +
                            basePath + clip.Name + ".wav" + "." + Environment.NewLine +
                            "The error message was: " + clip.LastErrorMessage, "Export Error",
                            MessageBoxButtons.OK, MessageBoxIcon.Error);
                    }
                    if (result == DialogResult.No)
                    {
                        Invoke(new MethodInvoker(DestroyProgressWindow));
                        return;
                    }
                }
                else
                {
                    if (_ExportLoopDataAsText)
                    {
                        WriteLoopDataToFile(clip, basePath + clip.Name + "-LoopPoints.txt");
                    }
                }
                clipNum++;
            }
            Invoke(new MethodInvoker(DestroyProgressWindow));
        }

        protected void WriteLoopDataToFile(BLCDDA.SoundClip clip, string path)
        {
            if (clip.LoopPoints == null)
            {
                // nothing to do
                return;
            }
            try
            {
                FileStream outStream = new FileStream(path, FileMode.Create, FileAccess.Write);
                StreamWriter outWriter = new StreamWriter(outStream);
                int loopNum = 1;
                foreach (BLDA.LoopPoint loop in clip.LoopPoints)
                {
                    outWriter.WriteLine("Loop " + string.Format("{0:00}", loopNum));
                    double startTime = ((double)loop.Start / (double)clip.SampleRate);
                    double endTime = ((double)loop.End / (double)clip.SampleRate);
                    outWriter.WriteLine("\tStart Sample: " + loop.Start);
                    outWriter.WriteLine("\tStart Time: " + BLUI.TimeDisplay.GetTimeString(startTime));
                    outWriter.WriteLine("\tEnd Sample: " + loop.End);
                    outWriter.WriteLine("\tEnd Time: " + BLUI.TimeDisplay.GetTimeString(endTime));
                    outWriter.WriteLine();
                    loopNum++;
                }
                outWriter.Close();
                outStream.Close();
            }
            catch (Exception ex)
            {
                MessageBox.Show("An error occurred while writing the text file " + path + Environment.NewLine +
                    "The error message was: " + clip.LastErrorMessage, "Export Error",
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        #endregion

        #region Utility functions related to the UI

        protected void EnablePlaybackControls(bool enabled)
        {
            btnPlay.Enabled = enabled;
            btnPause.Enabled = enabled;
            btnStop.Enabled = enabled;
        }

        protected void EnableExportControls(bool enabled)
        {
            currentSoundClipToolStripMenuItem.Enabled = enabled;
            btnExport.Enabled = enabled;
        }

        protected void EnableExportAllControls(bool enabled)
        {
            allSoundClipsToolStripMenuItem.Enabled = enabled;
            btnExportAll.Enabled = enabled;
        }

        protected void PlaySound()
        {
            if (_CurrentClip != null)
            {
                if ((_Timer != null) && (_Timer.Enabled))
                {
                    _Audio.CurrentPosition = (_Audio.Duration * _WaveformStartPercent);
                    _Audio.Play();
                }
                else
                {
                    _Audio = new DXAVP.Audio(_CurrentClip.WAVFile.TempFilePath, false);
                    _Audio.CurrentPosition = (_Audio.Duration * _WaveformStartPercent);
                    _Audio.Play();
                    //_Audio.Stopping += new EventHandler(DisposeAudio);
                    // reset loop system
                    _CurrentLoop = 0;
                    _LoopCount = 0;
                    CalculateLoopVariables();
 
                    _Timer = new System.Windows.Forms.Timer();
                    _Timer.Interval = 10;
                    _Timer.Tick += new EventHandler(_Timer_Tick);
                    _Timer.Start();
                }
            }
        }

        protected void PauseSound()
        {
            if ((_Audio != null) && (!_Audio.Disposed))
            {
                if (_Audio.Playing)
                {
                    _Audio.Pause();
                }
                // DirectSound Audio object lies about its state, have to check position too
                else if ((_Audio.Paused) && (_Audio.CurrentPosition > 0))
                {
                    _Audio.Play();
                }
            }
        }

        protected void StopSound()
        {
            if ((_Audio != null) && (!_Audio.Disposed))
            {
                _Audio.Stop();
                _Audio.CurrentPosition = 0;
                _Timer.Stop();
                _Audio.Dispose();
                _Audio = null;
            }
            _WaveformStartPercent = 0;
            _CurrentLoop = 0;
            _LoopCount = 0;
            UpdatePositionCursors();
        }

        protected void SeekInSound(double newPosition)
        {
            if ((_Audio != null) && (!_Audio.Disposed))
            {
                _Audio.CurrentPosition = (newPosition * _Audio.Duration);
            }
            else
            {
                _WaveformStartPercent = newPosition;
            }
        }

        protected void CreateInitialWaveformDisplay()
        {
            SetupTimeDisplay();
            BLDA.WaveformDisplay wfd = CreateWaveformDisplay(39);
            wfd.Dock = DockStyle.Fill;
            pnlWaveform.Controls.Add(wfd);
            _WaveformDisplays = new BLDA.WaveformDisplay[1] { wfd };
        }

        protected void SetupTimeDisplay()
        {
            TimeCounter.BackColor = this.BackColor;
            TimeCounter.ForeColor = this.ForeColor;
            //TimeCounter.Height = _TimeDisplayHeight;
            pnlTimeDisplay.Height = _TimeDisplayHeight;
            Font currentFont = TimeCounter.Font;
            TimeCounter.Font = new Font(currentFont.FontFamily, 18, currentFont.Style);
            //TimeCounter.Dock = DockStyle.Bottom;
            //pnlWaveform.Controls.Add(TimeCounter);
        }

        protected void DoPostFileLoadUISetup()
        {
            UpdateClipList();
            if (_CurrentFile != null)
            {
                if (_CurrentFile.SoundClips.GetLength(0) > 0)
                {
                    EnableExportControls(true);
                }
                if (_CurrentFile.SoundClips.GetLength(0) > 1)
                {
                    EnableExportAllControls(true);
                }
            }
        }

        protected void UpdateClipList()
        {
            lbClipList.Items.Clear();
            foreach (BLCDDA.SoundClip clip in _CurrentFile.SoundClips)
            {
                lbClipList.Items.Add(clip.Name);
            }
            lbClipList.SelectedIndex = 0;
        }

        protected void DisplayClipInfo()
        {
            txtClipInfo.Clear();
            Hashtable fileInfo = _CurrentFile.GetInfo();
            txtClipInfo.AppendText("File '" + GetFileNameFromPath(_CurrentFile.Path) + "'" + Environment.NewLine);
            foreach (string key in fileInfo.Keys)
            {
                txtClipInfo.AppendText(key.ToString() + ": " + fileInfo[key].ToString() + Environment.NewLine);
            }

            Hashtable clipInfo = _CurrentClip.GetInfo();
            txtClipInfo.AppendText(Environment.NewLine + "Clip '" + _CurrentClip.Name + "'" + Environment.NewLine);
            foreach (string key in clipInfo.Keys)
            {
                txtClipInfo.AppendText(key.ToString() + ": " + clipInfo[key].ToString() + Environment.NewLine);
            }
            this.Text = _CurrentClip.Name + " (" + GetFileNameFromPath(_CurrentFile.Path) + ") - The Sound of Her Wings";
        }

        protected void DisplayCurrentClipThreaded()
        {
            if (_DrawWaveform)
            {
                // write each channel of wav data to a temp file
                // read the channel data from the wav file into the temp files
                Invoke(new MethodInvoker(CreateProgressWindow));
                progressWindow.SetTitle("Loading");
                progressWindow.SetMessage("Extracting Channel Data");
                Thread copyThread = new Thread(CopyWaveDataToTempFiles);
                copyThread.Start();

                // wait for the loading to finish, and keep the load value in this class updated
                do
                {
                    progressWindow.SetProgress((int)(_Progress * (double)100));
                    Thread.Sleep(5);
                }
                while (copyThread.IsAlive);
            }
            Invoke(new MethodInvoker(DestroyProgressWindow));
            Invoke(new MethodInvoker(DisplayCurrentClip));
        }

        protected void DisplayCurrentClip()
        {
            // reset start time when a new clip is loaded
            _WaveformStartPercent = 0;

            pnlWaveform.Controls.Clear();

            // set number of gridlines based on how many waveforms are being displayed
            // to make sure they are spaced in a nice way
            int numGridLinesX;
            switch (_CurrentClip.NumChannels)
            {
                case 1:
                    numGridLinesX = 39;
                    break;
                case 2:
                    numGridLinesX = 19;
                    break;
                default:
                    numGridLinesX = 9;
                    break;
            }

            _WaveformDisplays = new BLDA.WaveformDisplay[_CurrentClip.NumChannels];
            for (int channelNum = 0; channelNum < _CurrentClip.NumChannels; channelNum++)
            {
                _WaveformDisplays[channelNum] = CreateWaveformDisplay(numGridLinesX);
                pnlWaveform.Controls.Add(_WaveformDisplays[channelNum]);
            }

            SetupTimeDisplay();

            ResizeAndRepositionWaveformDisplays();
            // now that they're sized correctly for the first time, have them load their data
            // this is done as a second iteration through the array to avoid two calls to the waveform drawing function
            for (int channelNum = 0; channelNum < _CurrentClip.NumChannels; channelNum++)
            {
                if (_DrawWaveform)
                {
                    try
                    {
                        _WaveformDisplays[channelNum].SetStream(_ChannelTempFilePaths[channelNum],
                            (int)_CurrentClip.WAVFile.SampleRate, _CurrentClip.WAVFile.BitsPerSample,
                            0, _CurrentClip.WAVFile.NumSamplesPerChannel);
                    }
                    catch (Exception ex)
                    {
                        MessageBox.Show("Error creating temporary files: " + ex.Message, "Error",
                            MessageBoxButtons.OK, MessageBoxIcon.Error);
                    }
                }
                _WaveformDisplays[channelNum].Invalidate();
            }

            // set the total time for the time display
            TimeCounter.SetTotalTime(((double)_CurrentClip.Length / 
                ((double)_CurrentClip.SampleRate * (double)(_CurrentClip.BitDepth / 8) * (double)_CurrentClip.NumChannels)) );

            // enable playback controls
            EnablePlaybackControls(true);
        }

        protected void CopyWaveDataToTempFiles()
        {
            // delete any previous temp files if there are any
            if (_ChannelTempFilePaths != null)
            {
                foreach (string path in _ChannelTempFilePaths)
                {
                    if (File.Exists(path))
                    {
                        try
                        {
                            File.Delete(path);
                        }
                        catch (Exception ex)
                        {
                            // do nothing
                        }
                    }
                }
            }
            try
            {
                FileStream inStream = new FileStream(_CurrentClip.WAVFile.TempFilePath, FileMode.Open, FileAccess.Read);
                inStream.Seek(_CurrentClip.WAVFile.HeaderLength, SeekOrigin.Begin);
                BinaryReader inReader = new BinaryReader(inStream);
                BinaryWriter[] outWriters = new BinaryWriter[_CurrentClip.NumChannels];
                _ChannelTempFilePaths = new string[_CurrentClip.NumChannels];
                for (int channelNum = 0; channelNum < _CurrentClip.NumChannels; channelNum++)
                {
                    _ChannelTempFilePaths[channelNum] = Path.GetTempFileName();
                    outWriters[channelNum] = new BinaryWriter(new FileStream(_ChannelTempFilePaths[channelNum], FileMode.Create, FileAccess.ReadWrite));
                }
                _Progress = 0;
                for (int sampleNum = 0; sampleNum < _CurrentClip.WAVFile.NumSamplesPerChannel; sampleNum++)
                {
                    _Progress = (double)sampleNum / (double)_CurrentClip.WAVFile.NumSamplesPerChannel;
                    for (int channelNum = 0; channelNum < _CurrentClip.NumChannels; channelNum++)
                    {
                        if (_CurrentClip.WAVFile.BitsPerSample == 8)
                        {
                            outWriters[channelNum].Write((byte)inReader.ReadByte());
                        }
                        else
                        {
                            outWriters[channelNum].Write(inReader.ReadInt16());
                        }
                    }
                }
                _Progress = 1;
                inStream.Close();
                for (int channelNum = 0; channelNum < _CurrentClip.NumChannels; channelNum++)
                {
                    outWriters[channelNum].Close();
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error creating temporary files: " + ex.Message, "Error",
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        protected BLDA.WaveformDisplay CreateWaveformDisplay(int numGridLinesX)
        {
            BLDA.WaveformDisplay wfd = new BLDA.WaveformDisplay();
            Font currentFont = wfd.Font;
            wfd.NumGridLinesX = numGridLinesX;
            wfd.NumGridLinesY = 39;

            wfd.BackgroundColour = _WFBackgroundColour;
            wfd.GridColour = _WFGridColour;
            wfd.WaveformColour = _WFWaveformColour;
            wfd.WaveformMedianColour = _WFWaveformMedianColour;
            wfd.XAxisColour = _WFXAxisColour;
            wfd.CursorColour = _WFCursorColour;
            wfd.DrawWaveform = _DrawWaveform;
            wfd.UseGradientDrawMode = _UseGradientDrawMode;

            wfd.Font = new Font(currentFont.FontFamily, 14, currentFont.Style);

            // bind to the click handler in this class
            wfd.WaveformClicked += new BenLincoln.DigitalAudio.WaveformDisplay.WaveformClickedDelegate(WaveformClick);

            return wfd;
        }

        protected void ResizeAndRepositionWaveformDisplays()
        {
            // abort if waveform displays don't exist yet
            if ((_WaveformDisplays == null) || (_CurrentClip == null))
            {
                return;
            }

            int wfSpacing = 2;  // number of points to space the displays vertically with in a multi-channel display

            int wfDisplayHeight;

            if (_CurrentClip.NumChannels == 1)
            {
                wfDisplayHeight = pnlWaveform.Height;
            }
            else
            {
                wfDisplayHeight = (pnlWaveform.Height - (wfSpacing * (_CurrentClip.NumChannels - 1))) / _CurrentClip.NumChannels;
            }

            int displayNum = 0;
            foreach (BLDA.WaveformDisplay wfd in _WaveformDisplays)
            {
                wfd.Location = new Point(0, ((wfDisplayHeight + wfSpacing) * displayNum));
                wfd.Size = new Size(pnlWaveform.Width, wfDisplayHeight);
                displayNum++;
            }
        }

        protected void UpdatePositionCursors()
        {
            if ((_Audio != null) && (!_Audio.Disposed))
            {
                double percentPlayed = (_Audio.CurrentPosition / _Audio.Duration);
                foreach (BLDA.WaveformDisplay wfd in _WaveformDisplays)
                {
                    wfd.CursorPosition = (float)percentPlayed;
                    wfd.Invalidate();
                }
                TimeCounter.SetElapsedPercent(percentPlayed);
            }
            else
            {
                //double percentPlayed = 0;
                foreach (BLDA.WaveformDisplay wfd in _WaveformDisplays)
                {
                    //wfd.CursorPosition = (float)percentPlayed;
                    wfd.CursorPosition = (float)_WaveformStartPercent;
                    wfd.Invalidate();
                }
                TimeCounter.SetElapsedPercent(_WaveformStartPercent);
            }
        }

        protected void CenterChildForm(Form childForm, Form parentForm)
        {
            // manually center the window in relation to its parent, since Windows doesn't do it correctly
            childForm.Location = new Point
            (
            parentForm.Location.X + ((parentForm.Width - childForm.Width) / 2),
            parentForm.Location.Y + ((parentForm.Height - childForm.Height) / 2)
            );
        }

        protected void CreateProgressWindow()
        {
            if (progressWindow != null)
            {
                progressWindow.Dispose();
            }
            progressWindow = new BLUI.ProgressWindow();
            progressWindow.Title = "Loading";
            progressWindow.SetMessage("Reading data");
            progressWindow.Icon = this.Icon;
            progressWindow.Owner = this;
            progressWindow.TopLevel = true;
            progressWindow.ShowInTaskbar = false;
            progressWindow.StartPosition = FormStartPosition.Manual;
            CenterChildForm(progressWindow, this);
            this.Enabled = false;
            progressWindow.Show();
        }

        protected void DestroyProgressWindow()
        {
            this.Enabled = true;
            progressWindow.Hide();
            progressWindow.Dispose();
        }

        protected void ShowPrefsWindow()
        {
            if (prefsWindow != null)
            {
                prefsWindow.Dispose();
            }
            prefsWindow = new frmPreferences();

            // set window pref values to variables from this class
            prefsWindow.DrawWaveform = _DrawWaveform;
            prefsWindow.UseGradientDrawMode = _UseGradientDrawMode;

            prefsWindow.WFBackgroundColour = _WFBackgroundColour.ToArgb();
            prefsWindow.WFGridColour = _WFGridColour.ToArgb();
            prefsWindow.WFXAxisColour = _WFXAxisColour.ToArgb();
            prefsWindow.WFWaveformColour = _WFWaveformColour.ToArgb();
            prefsWindow.WFWaveformMedianColour = _WFWaveformMedianColour.ToArgb();
            prefsWindow.WFCursorColour = _WFCursorColour.ToArgb();

            prefsWindow.LoopIterations = _LoopIterations;

            prefsWindow.ExportLoopDataAsText = _ExportLoopDataAsText;

            prefsWindow.StartPosition = FormStartPosition.Manual;
            CenterChildForm(prefsWindow, this);

            prefsWindow.ShowDialog();

            if (prefsWindow.Result == frmPreferences.PrefsResult.Save)
            {
                _DrawWaveform = prefsWindow.DrawWaveform;
                _UseGradientDrawMode = prefsWindow.UseGradientDrawMode;

                _WFBackgroundColour = Color.FromArgb(prefsWindow.WFBackgroundColour);
                _WFGridColour = Color.FromArgb(prefsWindow.WFGridColour);
                _WFXAxisColour = Color.FromArgb(prefsWindow.WFXAxisColour);
                _WFWaveformColour = Color.FromArgb(prefsWindow.WFWaveformColour);
                _WFWaveformMedianColour = Color.FromArgb(prefsWindow.WFWaveformMedianColour);
                _WFCursorColour = Color.FromArgb(prefsWindow.WFCursorColour);

                _LoopIterations = prefsWindow.LoopIterations;

                _ExportLoopDataAsText = prefsWindow.ExportLoopDataAsText;
                
                // write the new prefs to the registry
                WriteOptionsToRegistry();

                // recreate the waveform displays with the new settings
                //if (_CurrentClip == null)
                //{
                //    pnlWaveform.Controls.Clear();
                //    CreateInitialWaveformDisplay();
                //}
                //else
                //{
                //    DisplayCurrentClip();
                //}
                foreach (BLDA.WaveformDisplay wfd in pnlWaveform.Controls)
                {
                    wfd.DrawWaveform = _DrawWaveform;
                    wfd.UseGradientDrawMode = _UseGradientDrawMode;
                    wfd.BackgroundColour = _WFBackgroundColour;
                    wfd.GridColour = _WFGridColour;
                    wfd.XAxisColour = _WFXAxisColour;
                    wfd.WaveformColour = _WFWaveformColour;
                    wfd.WaveformMedianColour = _WFWaveformMedianColour;
                    wfd.CursorColour = _WFCursorColour;
                    wfd.RedrawWaveformImage();
                }
            }
        }

        protected void CreateToolTip(System.Windows.Forms.Control boundControl, string message)
        {
            ToolTip tTip = new ToolTip();
            tTip.AutoPopDelay = 1000;
            tTip.InitialDelay = 1000;
            tTip.ReshowDelay = 500;
            tTip.ShowAlways = true;
            tTip.SetToolTip(boundControl, message);
        }

        protected void ResizeClipList()
        {
            // workaround for the stupid list box buggy piece of crap fill dock mode
            lbClipList.Height = spltFileInfo.Height - (spltFileInfo.Panel2.Height + pnlClipListText.Height);
        }

        #endregion

        #region Utility functions for low-level work

        protected void CalculateLoopVariables()
        {
            // calculate loop point percentages if there are any
            if ((_CurrentClip.LoopPoints != null) && (_CurrentLoop < _CurrentClip.LoopPoints.Length))
            {
                long clipLengthInSamples = _CurrentClip.Length / ((_CurrentClip.BitDepth / 8) * _CurrentClip.NumChannels);
                _CurrentLoopStart = ((double)_CurrentClip.LoopPoints[_CurrentLoop].Start / (double)clipLengthInSamples);
                _CurrentLoopEnd = ((double)_CurrentClip.LoopPoints[_CurrentLoop].End / (double)clipLengthInSamples);
            }
        }

        protected void ReadOptionsFromRegistry()
        {
            // initialize any missing values
            _PrefManager.InitRegistryKeys();
            _PrefManager.InitRegistryDWordValues(GetCurrentDWordPrefs());
            _PrefManager.InitRegistryStringValues(GetCurrentStringPrefs());

            // read the values from the registry and use them
            SetVariablesFromDWordPrefs(_PrefManager.ReadDWordValuesFromRegistry());
            SetVariablesFromStringPrefs(_PrefManager.ReadStringValuesFromRegistry());
        }

        protected void WriteOptionsToRegistry()
        {
            WriteDWordPrefsToRegistry(GetCurrentDWordPrefs());
            WriteStringPrefsToRegistry(GetCurrentStringPrefs());
        }

        protected void SetVariablesFromDWordPrefs(ArrayList regValues)
        {
            foreach (BLS.RegistryDWordValue value in regValues)
            {
                switch (value.ValueName)
                {
                    case "Use Gradient Drawing Mode":
                        if (value.Value == 1)
                        {
                            _UseGradientDrawMode = true;
                        }
                        else
                        {
                            _UseGradientDrawMode = false;
                        }
                        break;
                    case "Draw Waveform":
                        if (value.Value == 1)
                        {
                            _DrawWaveform = true;
                        }
                        else
                        {
                            _DrawWaveform = false;
                        }
                        break;
                    case "WaveformBackgroundColour":
                        _WFBackgroundColour = Color.FromArgb((int)value.Value);
                        break;
                    case "WaveformGridColour":
                        _WFGridColour = Color.FromArgb((int)value.Value);
                        break;
                    case "WaveformXAxisColour":
                        _WFXAxisColour = Color.FromArgb((int)value.Value);
                        break;
                    case "WaveformWaveformColour":
                        _WFWaveformColour = Color.FromArgb((int)value.Value);
                        break;
                    case "WaveformMedianColour":
                        _WFWaveformMedianColour = Color.FromArgb((int)value.Value);
                        break;
                    case "WaveformPositionColour":
                        _WFCursorColour = Color.FromArgb((int)value.Value);
                        break;
                    case "Loop Iterations":
                        _LoopIterations = (int)value.Value;
                        break;
                    case "Export Loop Data As Text":
                        if (value.Value == 1)
                        {
                            _ExportLoopDataAsText = true;
                        }
                        else
                        {
                            _ExportLoopDataAsText = false;
                        }
                        break;
                }
            }
        }

        protected void SetVariablesFromStringPrefs(ArrayList regValues)
        {
            foreach (BLS.RegistryStringValue value in regValues)
            {
                switch (value.ValueName)
                {
                    case "Last Open Folder":
                        _LastOpenFolder = value.Value;
                        break;
                    case "Last Export Folder":
                        _LastExportFolder = value.Value;
                        break;
                    case "Last Export All Folder":
                        _LastExportAllFolder = value.Value;
                        break;
                }
            }
        }

        protected void WriteDWordPrefsToRegistry(ArrayList prefs)
        {
            _PrefManager.WriteDWordValuesToRegistry(GetCurrentDWordPrefs());
        }

        protected void WriteStringPrefsToRegistry(ArrayList prefs)
        {
            _PrefManager.WriteStringValuesToRegistry(GetCurrentStringPrefs());
        }

        // note: the "current" values are the defaults before the real current values have been loaded
        // from the registry at startup
        protected ArrayList GetCurrentDWordPrefs()
        {
            ArrayList prefs = new ArrayList();
            int useGradient = 1;
            if (!_UseGradientDrawMode)
            {
                useGradient = 0;
            }
            int drawWaveform = 1;
            if (!_DrawWaveform)
            {
                drawWaveform = 0;
            }
            int exportLoopDataAsText = 1;
            if (!_ExportLoopDataAsText)
            {
                exportLoopDataAsText = 0;
            }
            prefs.Add(new BLS.RegistryDWordValue("Use Gradient Drawing Mode", (int)useGradient));
            prefs.Add(new BLS.RegistryDWordValue("Draw Waveform", (int)drawWaveform));
            prefs.Add(new BLS.RegistryDWordValue("WaveformBackgroundColour", (int)_WFBackgroundColour.ToArgb()));
            prefs.Add(new BLS.RegistryDWordValue("WaveformGridColour", (int)_WFGridColour.ToArgb()));
            prefs.Add(new BLS.RegistryDWordValue("WaveformXAxisColour", (int)_WFXAxisColour.ToArgb()));
            prefs.Add(new BLS.RegistryDWordValue("WaveformWaveformColour", (int)_WFWaveformColour.ToArgb()));
            prefs.Add(new BLS.RegistryDWordValue("WaveformMedianColour", (int)_WFWaveformMedianColour.ToArgb()));
            prefs.Add(new BLS.RegistryDWordValue("WaveformPositionColour", (int)_WFCursorColour.ToArgb()));
            prefs.Add(new BLS.RegistryDWordValue("Loop Iterations", (int)_LoopIterations));
            prefs.Add(new BLS.RegistryDWordValue("Export Loop Data As Text", (int)exportLoopDataAsText));
            return prefs;
        }

        protected ArrayList GetCurrentStringPrefs()
        {
            ArrayList prefs = new ArrayList();
            prefs.Add(new BLS.RegistryStringValue("Last Open Folder", _LastOpenFolder));
            prefs.Add(new BLS.RegistryStringValue("Last Export Folder", _LastExportFolder));
            prefs.Add(new BLS.RegistryStringValue("Last Export All Folder", _LastExportAllFolder));
            return prefs;
        }

        protected string GetFileNameFromPath(string path)
        {
            string[] pathSplit = path.Split('\\');
            return pathSplit[pathSplit.GetUpperBound(0)];
        }

        protected string GetDirectoryNameFromPath(string path)
        {
            string[] pathSplit = path.Split('\\');
            string outPath = "";
            for (int i = 0; i < pathSplit.GetUpperBound(0); i++)
            {
                outPath += pathSplit[i] + "\\";
            }
            return outPath;
        }

        protected void DeleteTempFiles()
        {
            if (_CurrentFile != null)
            {
                _CurrentFile.Dispose();
            }
            if (_ChannelTempFilePaths != null)
            {
                foreach (string tFile in _ChannelTempFilePaths)
                {
                    if (tFile != null)
                    {
                        try
                        {
                            File.Delete(tFile);
                        }
                        catch (IOException ex)
                        {
                            // do nothing
                        }
                    }
                }
            }
        }

        #endregion

        #region Event handlers for the UI

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        private void openToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OpenFile();
        }

        private void optionsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ShowPrefsWindow();
        }

        private void currentSoundClipToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ExportCurrent();
        }

        private void allSoundClipsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ExportAll();
        }

        private void aboutTheSoundOfHerWingsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (aboutWindow != null)
            {
                aboutWindow.Dispose();
            }
            aboutWindow = new frmAbout();
            aboutWindow.StartPosition = FormStartPosition.Manual;
            CenterChildForm(aboutWindow, this);
            aboutWindow.Show();
        }

        private void btnOpen_Click(object sender, EventArgs e)
        {
            OpenFile();
        }

        private void btnExport_Click(object sender, EventArgs e)
        {
            ExportCurrent();
        }

        private void btnExportAll_Click(object sender, EventArgs e)
        {
            ExportAll();
        }

        private void ClipList_SelectedIndexChanged(object sender, EventArgs e)
        {
            if ((_CurrentFile != null) && (lbClipList.SelectedIndex > -1) && (lbClipList.SelectedIndex != _LastClipListIndex))
            {
                // update the last index
                _LastClipListIndex = lbClipList.SelectedIndex;
                // if there is a sound playing, stop it
                StopSound();
                _CurrentClip = _CurrentFile.SoundClips[lbClipList.SelectedIndex];
                DisplayClipInfo();
                Thread displayThread = new Thread(DisplayCurrentClipThreaded);
                displayThread.Start();
                //DisplayCurrentClip();
            }
        }

        private void btnPlay_Click(object sender, EventArgs e)
        {
            PlaySound();   
        }

        private void btnPause_Click(object sender, EventArgs e)
        {
            PauseSound();
        }

        private void btnStop_Click(object sender, EventArgs e)
        {
            StopSound();
        }

        protected void WaveformClick(object sender, EventArgs e)
        {
            BLDA.WaveformDisplay wfd = (BLDA.WaveformDisplay)sender;
            if (!wfd.IsLoadingData && !wfd.IsDrawing)
            {
                SeekInSound(((double)wfd.LastClickPosition.X / (double)wfd.Width));
            }
            UpdatePositionCursors();
        }

        protected override void OnResize(EventArgs e)
        {
            base.OnResize(e);

            ResizeClipList();

            ResizeAndRepositionWaveformDisplays();
        }

        private void clipListResize(object sender, EventArgs e)
        {
            ResizeClipList();
        }

        private void spltFileInfo_SplitterMoved(object sender, SplitterEventArgs e)
        {
            ResizeClipList();
        }

        #endregion

        #region Event handlers for other events

        protected void DisposeAudio(Object obj, EventArgs e)
        {
            _Audio.Dispose();
            _Audio = null;
        }

        void _Timer_Tick(object sender, EventArgs e)
        {
            // handle loops if there are any
            if ((_CurrentClip.LoopPoints != null) && (_LoopIterations > 1))
            {
                double audioPercent = _Audio.CurrentPosition / _Audio.Duration;
                if ((_CurrentLoop < _CurrentClip.LoopPoints.Length) && (audioPercent >= _CurrentLoopEnd))
                {
                    _LoopCount++;
                    if (_LoopCount >= _LoopIterations)
                    {
                        _CurrentLoop++;
                        _LoopCount = 0;
                        CalculateLoopVariables();
                    }
                    else
                    {
                        SeekInSound(_CurrentLoopStart);
                    }
                }
            }
            if (_Audio.CurrentPosition == _Audio.Duration)
            {
                _Timer.Stop();
                _Audio.Dispose();
            }
            UpdatePositionCursors();
        }

        #endregion

        #region Interface methods

        public double Progress()
        {
            return _Progress;
        }

        #endregion

    }
}