// The Sound of Her Wings
// Copyright 2006-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of The Sound of Her Wings.

// The Sound of Her Wings is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// The Sound of Her Wings is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with The Sound of Her Wings (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using BLS = BenLincoln.Shared;

namespace TheSoundOfHerWings
{
    public partial class frmPreferences : Form
    {
        public enum PrefsResult
        {
            Discard,
            Save
        }

        protected struct ColourScheme
        {
            public string Name;
            public int BackgroundColour;
            public int GridColour;
            public int XAxisColour;
            public int WaveformColour;
            public int WaveformMedianColour;
            public int CursorColour;
        }

        // preset colour schemes
        protected ColourScheme[] _PresetSchemes;

        // what button did the user click?
        protected PrefsResult _Result;

        // what state are the controls in?
        protected bool _DrawWaveform = true;
        protected bool _UseGradientDrawMode = true;

        protected int _WFBackgroundColour = unchecked((int)0xff808080);
        protected int _WFGridColour = unchecked((int)0x80404040);
        protected int _WFXAxisColour = unchecked((int)0x40dcdcdc);
        protected int _WFWaveformColour = unchecked((int)0xa06400ff);
        protected int _WFWaveformMedianColour = unchecked((int)0xbeff18ff);
        protected int _WFCursorColour = unchecked((int)0x801818ff);

        protected int _LoopIterations = 1;

        protected bool _ExportLoopDataAsText = false;

        #region Properties

        public PrefsResult Result
        {
            get
            {
                return _Result;
            }
        }

        public bool DrawWaveform
        {
            get
            {
                return _DrawWaveform;
            }
            set
            {
                _DrawWaveform = value;
            }
        }

        public bool UseGradientDrawMode
        {
            get
            {
                return _UseGradientDrawMode;
            }
            set
            {
                _UseGradientDrawMode = value;
            }
        }

        public int WFBackgroundColour
        {
            get
            {
                return _WFBackgroundColour;
            }
            set
            {
                _WFBackgroundColour = value;
            }
        }

        public int WFGridColour
        {
            get
            {
                return _WFGridColour;
            }
            set
            {
                _WFGridColour = value;
            }
        }

        public int WFXAxisColour
        {
            get
            {
                return _WFXAxisColour;
            }
            set
            {
                _WFXAxisColour = value;
            }
        }

        public int WFWaveformColour
        {
            get
            {
                return _WFWaveformColour;
            }
            set
            {
                _WFWaveformColour = value;
            }
        }

        public int WFWaveformMedianColour
        {
            get
            {
                return _WFWaveformMedianColour;
            }
            set
            {
                _WFWaveformMedianColour = value;
            }
        }

        public int WFCursorColour
        {
            get
            {
                return _WFCursorColour;
            }
            set
            {
                _WFCursorColour = value;
            }
        }

        public int LoopIterations
        {
            get
            {
                return _LoopIterations;
            }
            set
            {
                _LoopIterations = value;
            }
        }

        public bool ExportLoopDataAsText
        {
            get
            {
                return _ExportLoopDataAsText;
            }
            set
            {
                _ExportLoopDataAsText = value;
            }
        }

        #endregion

        public frmPreferences()
        {
            InitializeComponent();

            SetupColourSchemes();
        }

        private void frmPreferences_Load(object sender, EventArgs e)
        {
            SettingsToUI();
        }

        private void btnSave_Click(object sender, EventArgs e)
        {
            try
            {
                UIToSettings();
                _Result = PrefsResult.Save;
                this.Hide();
            }
            catch (BLS.InvalidParameterValueException ex)
            {
                // do nothing, user needs to correct values
            }
        }

        private void btnDiscard_Click(object sender, EventArgs e)
        {
            _Result = PrefsResult.Discard;
            this.Hide();
        }

        // updates the UI with the current variable settings
        protected void SettingsToUI()
        {
            cbDrawWaveform.Checked = _DrawWaveform;
            cbUseGradientDraw.Checked = _UseGradientDrawMode;

            txtWFBackgroundColour.Text = string.Format("{0:X8}", _WFBackgroundColour);
            txtWFGridColour.Text = string.Format("{0:X8}", _WFGridColour);
            txtWFXAxisColour.Text = string.Format("{0:X8}", _WFXAxisColour);
            txtWFWaveformColour.Text = string.Format("{0:X8}", _WFWaveformColour);
            txtWFWaveformMedianColour.Text = string.Format("{0:X8}", _WFWaveformMedianColour);
            txtWFCursorColour.Text = string.Format("{0:X8}", _WFCursorColour);

            txtLoopIterations.Text = _LoopIterations.ToString();

            cbExportLoopData.Checked = _ExportLoopDataAsText;
        }

        // updates the variables with the UI settings
        protected void UIToSettings()
        {
            _DrawWaveform = cbDrawWaveform.Checked;
            _UseGradientDrawMode = cbUseGradientDraw.Checked;

            _WFBackgroundColour = ValidateHexNum(txtWFBackgroundColour, "Waveform Background", _WFBackgroundColour);
            _WFGridColour = ValidateHexNum(txtWFGridColour, "Waveform Grid", _WFGridColour);
            _WFXAxisColour = ValidateHexNum(txtWFXAxisColour, "Waveform Background", _WFXAxisColour);
            _WFWaveformColour = ValidateHexNum(txtWFWaveformColour, "Waveform Background", _WFWaveformColour);
            _WFWaveformMedianColour = ValidateHexNum(txtWFWaveformMedianColour, "Waveform Background", _WFWaveformMedianColour);
            _WFCursorColour = ValidateHexNum(txtWFCursorColour, "Waveform Background", _WFCursorColour);

            int loopIter = ValidateDecNum(txtLoopIterations, "Loop Iterations", _LoopIterations);
            if (loopIter > 0)
            {
                _LoopIterations = loopIter;
            }
            else
            {
                MessageBox.Show("Loops must be played at least once.", "Invalid Entry", 
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
                _LoopIterations = 1;
                txtLoopIterations.Text = "1";
            }

            _ExportLoopDataAsText = cbExportLoopData.Checked;
        }

        protected int ValidateHexNum(TextBox source, string fieldName, int originalValue)
        {
            int retVal = originalValue;
            try
            {
                retVal = unchecked((int)Int64.Parse(source.Text.Trim(), System.Globalization.NumberStyles.HexNumber));
            }
            catch (FormatException ex)
            {
                MessageBox.Show("The value you entered in the field " + fieldName +
                    " is not a valid 32-bit hexadecimal number.", "Invalid Entry",
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
                source.Text = string.Format("{0:X8}", originalValue);
                throw new BLS.InvalidParameterValueException("Invalid entry in field " + fieldName);
            }
            return retVal;
        }

        protected int ValidateDecNum(TextBox source, string fieldName, int originalValue)
        {
            int retVal = originalValue;
            try
            {
                retVal = (int)UInt16.Parse(source.Text.Trim());
            }
            catch (FormatException ex)
            {
                MessageBox.Show("The value you entered in the field " + fieldName +
                    " is not a valid positive number.", "Invalid Entry",
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
                source.Text = originalValue.ToString();
                throw new BLS.InvalidParameterValueException("Invalid entry in field " + fieldName);
            }
            return retVal;
        }

        // purple / black
        //Color _WFWaveformColour = Color.FromArgb(160, 0, 0, 0);
        //Color _WFWaveformMedianColour = Color.FromArgb(190, 255, 0, 255);

        // red / black
        //Color _WFWaveformColour = Color.FromArgb(160, 0, 0, 0);
        //Color _WFWaveformMedianColour = Color.FromArgb(190, 255, 0, 0);

        // green
        //Color _WFWaveformColour = Color.FromArgb(160, 0, 100, 0);
        //Color _WFWaveformMedianColour = Color.FromArgb(190, 30, 255, 30);

        // yellow-green
        //Color _WFWaveformColour = Color.FromArgb(160, 0, 40, 0);
        //Color _WFWaveformMedianColour = Color.FromArgb(190, 128, 255, 0);

        // green / black
        //Color _WFWaveformColour = Color.FromArgb(160, 0, 0, 0);
        //Color _WFWaveformMedianColour = Color.FromArgb(190, 0, 255, 0);

        //// white / black
        //Color _WFWaveformColour = Color.FromArgb(160, 0, 0, 0);
        //Color _WFWaveformMedianColour = Color.FromArgb(190, 255, 255, 255);

        // white / blue
        //Color _WFWaveformColour = Color.FromArgb(160, 0, 0, 255);
        //Color _WFWaveformMedianColour = Color.FromArgb(190, 255, 255, 255);

        // white / steel-blue
        //Color _WFWaveformColour = Color.FromArgb(160, 0, 0, 128);
        //Color _WFWaveformMedianColour = Color.FromArgb(190, 255, 255, 255);

        // yellow / red
        //Color _WFWaveformColour = Color.FromArgb(160, 255, 0, 0);
        //Color _WFWaveformMedianColour = Color.FromArgb(190, 255, 255, 0);

        // cockroach
        //Color _WFWaveformColour = Color.FromArgb(160, 128, 0, 0);
        //Color _WFWaveformMedianColour = Color.FromArgb(190, 255, 255, 0);

        // green / blue
        //Color _WFWaveformColour = Color.FromArgb(160, 0, 0, 255);
        //Color _WFWaveformMedianColour = Color.FromArgb(190, 0, 255, 0);

        // red / purple
        //Color _WFWaveformColour = Color.FromArgb(160, 255, 0, 255);
        //Color _WFWaveformMedianColour = Color.FromArgb(190, 255, 0, 0);

        // red / dark purple
        //Color _WFWaveformColour = Color.FromArgb(160, 128, 0, 255);
        //Color _WFWaveformMedianColour = Color.FromArgb(190, 255, 0, 0);

        // blood-red
        //Color _WFWaveformColour = Color.FromArgb(160, 96, 0, 0);
        //Color _WFWaveformMedianColour = Color.FromArgb(190, 255, 0, 0);

        // cotton candy
        //Color _WFWaveformColour = Color.FromArgb(160, 255, 200, 200);
        //Color _WFWaveformMedianColour = Color.FromArgb(190, 255, 255, 255);

        // white / pink
        //Color _WFWaveformColour = Color.FromArgb(160, 255, 0, 156);
        //Color _WFWaveformMedianColour = Color.FromArgb(190, 255, 255, 255);

        // green / dark purple
        //Color _WFWaveformColour = Color.FromArgb(160, 156, 0, 156);
        //Color _WFWaveformMedianColour = Color.FromArgb(190, 0, 255, 0);

        // green / purple
        //Color _WFWaveformColour = Color.FromArgb(160, 220, 0, 255);
        //Color _WFWaveformMedianColour = Color.FromArgb(190, 0, 255, 0);

        // gunmetal
        //Color _WFWaveformColour = Color.FromArgb(160, 146, 138, 113);
        //Color _WFWaveformMedianColour = Color.FromArgb(190, 243, 243, 243);

        // gunmetal 2
        //Color _WFWaveformColour = Color.FromArgb(160, 72, 68, 54);
        //Color _WFWaveformMedianColour = Color.FromArgb(190, 243, 243, 243);

        //// gunmetal 3
        //Color _WFWaveformColour = Color.FromArgb(160, 71, 64, 43);
        //Color _WFWaveformMedianColour = Color.FromArgb(190, 225, 238, 255);

        // gunmetal 4
        //Color _WFWaveformColour = Color.FromArgb(160, 71, 64, 43);
        //Color _WFWaveformMedianColour = Color.FromArgb(190, 238, 246, 255);

        // super-80s purple
        //Color _WFWaveformColour = Color.FromArgb(160, 156, 0, 255);
        //Color _WFWaveformMedianColour = Color.FromArgb(190, 255, 0, 255);

        protected void SetupColourSchemes()
        {
            ColourScheme pinkWhite = new ColourScheme();
            pinkWhite.Name = "80s Pink/White";
            pinkWhite.BackgroundColour = unchecked((int)0xff808080);
            pinkWhite.GridColour = unchecked((int)0x80404040);
            pinkWhite.XAxisColour = unchecked((int)0x40dcdcdc);
            pinkWhite.WaveformColour = unchecked((int)0x80ff009c);
            pinkWhite.WaveformMedianColour = unchecked((int)0xbeffffff);
            pinkWhite.CursorColour = unchecked((int)0x80ff009c);

            ColourScheme amplifiedReaver = new ColourScheme();
            amplifiedReaver.Name = "Amplified Reaver";
            amplifiedReaver.BackgroundColour = unchecked((int)0xff808080);
            amplifiedReaver.GridColour = unchecked((int)0x80404040);
            amplifiedReaver.XAxisColour = unchecked((int)0x40dcdcdc);
            amplifiedReaver.WaveformColour = unchecked((int)0xa0000000);
            amplifiedReaver.WaveformMedianColour = unchecked((int)0xbeff0000);
            amplifiedReaver.CursorColour = unchecked((int)0x80180000);

            ColourScheme bloodRed = new ColourScheme();
            bloodRed.Name = "Blood Red";
            bloodRed.BackgroundColour = unchecked((int)0xff808080);
            bloodRed.GridColour = unchecked((int)0x80404040);
            bloodRed.XAxisColour = unchecked((int)0x40dcdcdc);
            bloodRed.WaveformColour = unchecked((int)0xa0600000);
            bloodRed.WaveformMedianColour = unchecked((int)0xbeff0000);
            bloodRed.CursorColour = unchecked((int)0x80800000);

            ColourScheme cockroach = new ColourScheme();
            cockroach.Name = "Cockroach";
            cockroach.BackgroundColour = unchecked((int)0xff808080);
            cockroach.GridColour = unchecked((int)0x80404040);
            cockroach.XAxisColour = unchecked((int)0x40dcdcdc);
            cockroach.WaveformColour = unchecked((int)0xa0800000);
            cockroach.WaveformMedianColour = unchecked((int)0xbeffff00);
            cockroach.CursorColour = unchecked((int)0x80800000);

            ColourScheme cottonCandy = new ColourScheme();
            cottonCandy.Name = "Cotton Candy";
            cottonCandy.BackgroundColour = unchecked((int)0xff808080);
            cottonCandy.GridColour = unchecked((int)0x80404040);
            cottonCandy.XAxisColour = unchecked((int)0x40dcdcdc);
            cottonCandy.WaveformColour = unchecked((int)0xa0ffc8c8);
            cottonCandy.WaveformMedianColour = unchecked((int)0xbeffffff);
            cottonCandy.CursorColour = unchecked((int)0x80ff009c);

            ColourScheme darkReaver = new ColourScheme();
            darkReaver.Name = "Dark Reaver";
            darkReaver.BackgroundColour = unchecked((int)0xff808080);
            darkReaver.GridColour = unchecked((int)0x80404040);
            darkReaver.XAxisColour = unchecked((int)0x40dcdcdc);
            darkReaver.WaveformColour = unchecked((int)0xa0000000);
            darkReaver.WaveformMedianColour = unchecked((int)0xbeff00ff);
            darkReaver.CursorColour = unchecked((int)0x801818ff);

            ColourScheme green = new ColourScheme();
            green.Name = "Green";
            green.BackgroundColour = unchecked((int)0xff808080);
            green.GridColour = unchecked((int)0x80404040);
            green.XAxisColour = unchecked((int)0x40dcdcdc);
            green.WaveformColour = unchecked((int)0xa0006400);
            green.WaveformMedianColour = unchecked((int)0xbe1eff1e);
            green.CursorColour = unchecked((int)0x8000ff00);

            ColourScheme greenBlack = new ColourScheme();
            greenBlack.Name = "Green/Black";
            greenBlack.BackgroundColour = unchecked((int)0xff808080);
            greenBlack.GridColour = unchecked((int)0x80404040);
            greenBlack.XAxisColour = unchecked((int)0x40dcdcdc);
            greenBlack.WaveformColour = unchecked((int)0xa0000000);
            greenBlack.WaveformMedianColour = unchecked((int)0xbe00ff00);
            greenBlack.CursorColour = unchecked((int)0x8000ff00);

            ColourScheme greenBlue = new ColourScheme();
            greenBlue.Name = "Green/Blue";
            greenBlue.BackgroundColour = unchecked((int)0xff808080);
            greenBlue.GridColour = unchecked((int)0x80404040);
            greenBlue.XAxisColour = unchecked((int)0x40dcdcdc);
            greenBlue.WaveformColour = unchecked((int)0xa00000ff);
            greenBlue.WaveformMedianColour = unchecked((int)0xbe00ff00);
            greenBlue.CursorColour = unchecked((int)0x8000ffff);

            ColourScheme greenDarkPurple = new ColourScheme();
            greenDarkPurple.Name = "Green/Dark Purple";
            greenDarkPurple.BackgroundColour = unchecked((int)0xff808080);
            greenDarkPurple.GridColour = unchecked((int)0x80404040);
            greenDarkPurple.XAxisColour = unchecked((int)0x40dcdcdc);
            greenDarkPurple.WaveformColour = unchecked((int)0x809c009c);
            greenDarkPurple.WaveformMedianColour = unchecked((int)0xbe00ff00);
            greenDarkPurple.CursorColour = unchecked((int)0x808c008c);

            ColourScheme greenPurple = new ColourScheme();
            greenPurple.Name = "Green/Purple";
            greenPurple.BackgroundColour = unchecked((int)0xff808080);
            greenPurple.GridColour = unchecked((int)0x80404040);
            greenPurple.XAxisColour = unchecked((int)0x40dcdcdc);
            greenPurple.WaveformColour = unchecked((int)0x80dc00ff);
            greenPurple.WaveformMedianColour = unchecked((int)0xbe00ff00);
            greenPurple.CursorColour = unchecked((int)0x809c009c);

            ColourScheme gunmetal = new ColourScheme();
            gunmetal.Name = "Gunmetal";
            gunmetal.BackgroundColour = unchecked((int)0xff808080);
            gunmetal.GridColour = unchecked((int)0x80404040);
            gunmetal.XAxisColour = unchecked((int)0x40dcdcdc);
            gunmetal.WaveformColour = unchecked((int)0x8047402b);
            gunmetal.WaveformMedianColour = unchecked((int)0xbeeef6ff);
            gunmetal.CursorColour = unchecked((int)0x8047402b);

            ColourScheme lilibat = new ColourScheme();
            lilibat.Name = "Lilibat's Hair";
            lilibat.BackgroundColour = unchecked((int)0xff808080);
            lilibat.GridColour = unchecked((int)0x80404040);
            lilibat.XAxisColour = unchecked((int)0x40dcdcdc);
            //lilibat.WaveformColour = unchecked((int)0xa0dc00ff);
            lilibat.WaveformColour = unchecked((int)0xa0c000dc);
            lilibat.WaveformMedianColour = unchecked((int)0xbe1818ff);
            lilibat.CursorColour = unchecked((int)0x801818ff);

            ColourScheme defaultPurple = new ColourScheme();
            defaultPurple.Name = "Purple/Violet (Default)";
            defaultPurple.BackgroundColour = unchecked((int)0xff808080);
            defaultPurple.GridColour = unchecked((int)0x80404040);
            defaultPurple.XAxisColour = unchecked((int)0x40dcdcdc);
            defaultPurple.WaveformColour = unchecked((int)0xa06400ff);
            defaultPurple.WaveformMedianColour = unchecked((int)0xbeff18ff);
            defaultPurple.CursorColour = unchecked((int)0x801818ff);

            ColourScheme redDarkPurple = new ColourScheme();
            redDarkPurple.Name = "Red/Dark Purple";
            redDarkPurple.BackgroundColour = unchecked((int)0xff808080);
            redDarkPurple.GridColour = unchecked((int)0x80404040);
            redDarkPurple.XAxisColour = unchecked((int)0x40dcdcdc);
            redDarkPurple.WaveformColour = unchecked((int)0xa08000ff);
            redDarkPurple.WaveformMedianColour = unchecked((int)0xbeff0000);
            redDarkPurple.CursorColour = unchecked((int)0x80202020);

            ColourScheme redPurple = new ColourScheme();
            redPurple.Name = "Red/Purple";
            redPurple.BackgroundColour = unchecked((int)0xff808080);
            redPurple.GridColour = unchecked((int)0x80404040);
            redPurple.XAxisColour = unchecked((int)0x40dcdcdc);
            redPurple.WaveformColour = unchecked((int)0xa0ff00ff);
            redPurple.WaveformMedianColour = unchecked((int)0xbeff0000);
            redPurple.CursorColour = unchecked((int)0x80202020);

            ColourScheme steelBlue = new ColourScheme();
            steelBlue.Name = "Steel Blue";
            steelBlue.BackgroundColour = unchecked((int)0xff808080);
            steelBlue.GridColour = unchecked((int)0x80404040);
            steelBlue.XAxisColour = unchecked((int)0x40dcdcdc);
            steelBlue.WaveformColour = unchecked((int)0xa0000080);
            steelBlue.WaveformMedianColour = unchecked((int)0xbeffffff);
            steelBlue.CursorColour = unchecked((int)0x801818ff);

            ColourScheme whiteBlack = new ColourScheme();
            whiteBlack.Name = "White/Black";
            whiteBlack.BackgroundColour = unchecked((int)0xff808080);
            whiteBlack.GridColour = unchecked((int)0x80404040);
            whiteBlack.XAxisColour = unchecked((int)0x40dcdcdc);
            whiteBlack.WaveformColour = unchecked((int)0xa0000000);
            whiteBlack.WaveformMedianColour = unchecked((int)0xbeffffff);
            whiteBlack.CursorColour = unchecked((int)0x80bebebe);

            ColourScheme whiteBlue = new ColourScheme();
            whiteBlue.Name = "White/Blue";
            whiteBlue.BackgroundColour = unchecked((int)0xff808080);
            whiteBlue.GridColour = unchecked((int)0x80404040);
            whiteBlue.XAxisColour = unchecked((int)0x40dcdcdc);
            whiteBlue.WaveformColour = unchecked((int)0xa00000ff);
            whiteBlue.WaveformMedianColour = unchecked((int)0xbeffffff);
            whiteBlue.CursorColour = unchecked((int)0x801818ff);

            ColourScheme yellowGreen = new ColourScheme();
            yellowGreen.Name = "Yellow-Green";
            yellowGreen.BackgroundColour = unchecked((int)0xff808080);
            yellowGreen.GridColour = unchecked((int)0x80404040);
            yellowGreen.XAxisColour = unchecked((int)0x40dcdcdc);
            yellowGreen.WaveformColour = unchecked((int)0xa0002800);
            yellowGreen.WaveformMedianColour = unchecked((int)0xbe80ff00);
            yellowGreen.CursorColour = unchecked((int)0x80e0ff00);

            ColourScheme yellowRed = new ColourScheme();
            yellowRed.Name = "Yellow/Red";
            yellowRed.BackgroundColour = unchecked((int)0xff808080);
            yellowRed.GridColour = unchecked((int)0x80404040);
            yellowRed.XAxisColour = unchecked((int)0x40dcdcdc);
            yellowRed.WaveformColour = unchecked((int)0xa0ff0000);
            yellowRed.WaveformMedianColour = unchecked((int)0xbeffff00);
            yellowRed.CursorColour = unchecked((int)0x80ff0000);

            _PresetSchemes = new ColourScheme[]
            {
                pinkWhite, amplifiedReaver, bloodRed, cockroach, cottonCandy, darkReaver, green, 
                greenBlack, greenBlue, greenDarkPurple, greenPurple, gunmetal, lilibat, defaultPurple, 
                redDarkPurple, redPurple, steelBlue, whiteBlack, whiteBlue, yellowGreen, yellowRed
            };

            ddColourScheme.Items.Clear();

            foreach (ColourScheme scheme in _PresetSchemes)
            {
                ddColourScheme.Items.Add(scheme.Name);
            }

        }

        private void ddColourScheme_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (ddColourScheme.SelectedIndex == -1)
            {
                return;
            }

            ColourScheme scheme = _PresetSchemes[ddColourScheme.SelectedIndex];

            txtWFBackgroundColour.Text = string.Format("{0:X8}", scheme.BackgroundColour);
            txtWFGridColour.Text = string.Format("{0:X8}", scheme.GridColour);
            txtWFXAxisColour.Text = string.Format("{0:X8}", scheme.XAxisColour);
            txtWFWaveformColour.Text = string.Format("{0:X8}", scheme.WaveformColour);
            txtWFWaveformMedianColour.Text = string.Format("{0:X8}", scheme.WaveformMedianColour);
            txtWFCursorColour.Text = string.Format("{0:X8}", scheme.CursorColour);
        }

    }
}